      SUBROUTINE GMRES(N,M,ITERMAX,TOL,VWORK,LDVWORK,HWORK,
     &                 REVCOM,CLX,CLY,XCOM,IER)
C
C SOLVE Ax=b USING THE PRECONDITIONNED GMRES(M) METHOD
C
C ITERMAX:    ON ENTRY: MAXIMUM NUMBER OF ITERATIONS
C             ON EXIT : NUMBER OF MATRIX-VECTOR PRODUCTS
C
C TOL:        ON ENTRY: TOLERANCE FOR ITERATION
C             ON EXIT : NORM OF THE RESIDUAL ERROR
C
C VWORK:      VWORK(:,1): ON EXIT : CONTAINS THE APPROXIMATE SOLUTION x
C             VWORK(:,2): ON ENTRY: CONTAINS THE RIGHT-HAND SIDE b
C             VWORK(:,3:M+3): CONTAINS THE KRYLOV BASIS
C             VWORK(:,M+4): USED FOR REVERSE COMMUNICATION
C             ON EXIT VWORK(:,M+4) CONTAINS Dx
C
C HWORK:      HWORK(:,1:M): ON EXIT : CONTAINS THE UPPER HESSENBERG MATRIX
C             HWORK(:,M+1): ON EXIT : CONTAINS ||r_0||e_1
C             HWORK(:,M+2): ON EXIT : CONTAINS THE COSINES OF GIVENS ROTATIONS
C             HWORK(:,M+3): ON EXIT : CONTAINS THE   SINES OF GIVENS ROTATIONS
C
C REVCOM:     ON ENTRY REVCOM =  0
C             ON EXIT  REVCOM =  1 MATRIX VECTOR PRODUCT BY A
C                      REVCOM =  2 MATRIX VECTOR PRODUCT BY P^{-1}
C                      REVCOM =  3 MATRIX VECTOR PRODUCT BY D
C                      REVCOM =  4 INNER PRODUCT
C                      REVCOM =  5 MATRIX VECTOR PRODUCT BY D^{-1}
C                      REVCOM = -1 EXIT
C
C IER: INTEGER (OUTPUT) IER =  1 SUCCESSFUL CONVERGENCE
C                       IER =  0 NO CONVERGENCE AFTER ITERMAX ITERATIONS
C
C CAUTION: ONLY THE INPUT ONES CAN BE MODIFIED DURING OUT OF CORE COMPUTATIONS
C
C-----------------------------------------------------------------------
C    DECLARATIONS
C-----------------------------------------------------------------------
      IMPLICIT  NONE
C    SUBROUTINE VARIABLES
      INTEGER, INTENT(IN   ) :: N      ! SIZE OF THE PROBLEM
      INTEGER, INTENT(IN   ) :: M      ! SIZE OF THE RESTART
      INTEGER, INTENT(INOUT) :: ITERMAX
      REAL   , INTENT(INOUT) :: TOL
      INTEGER, INTENT(IN   ) :: LDVWORK ! LEADING DIMENSION OF VWORK, N IN GENERAL
      REAL   , INTENT(INOUT) :: VWORK(LDVWORK,M+4)
      REAL   , INTENT(INOUT) :: HWORK(M+1,M+3)
      INTEGER, INTENT(INOUT) :: REVCOM
      INTEGER, INTENT(  OUT) :: CLX    ! COLUMN OF  INPUT IN VWORK FOR REVERSE COMMUNICATION
      INTEGER, INTENT(  OUT) :: CLY    ! COLUMN OF OUTPUT IN VWORK FOR REVERSE COMMUNICATION
      REAL   , INTENT(INOUT) :: XCOM   ! ON ENTRY CONTAINS INNER PRODUCT WHEN REVCOM = 4
      INTEGER, INTENT(  OUT) :: IER
C    LOCAL VARIABLES
      INTEGER                :: K
      REAL                   :: NORMR,TEMP
C    LOCAL SAVED VARIABLES
      INTEGER, SAVE          :: I,ICOM,ITER,KMGS
      REAL   , SAVE          :: ERROR,NORMR0
C-----------------------------------------------------------------------
C    BEGINNING OF SUBROUTINE GMRES
C-----------------------------------------------------------------------
      IF (REVCOM==0) ICOM = REVCOM
      IF     (ICOM== 1) THEN
         GOTO 10
      ELSEIF (ICOM== 2) THEN
         GOTO 20
      ELSEIF (ICOM== 3) THEN
	 GOTO 30
      ELSEIF (ICOM== 4) THEN
	 GOTO 40
      ELSEIF (ICOM== 5) THEN
	 GOTO 50
      ELSEIF (ICOM== 6) THEN
	 GOTO 60
      ELSEIF (ICOM== 7) THEN
	 GOTO 70
      ELSEIF (ICOM== 8) THEN
	 GOTO 80
      ELSEIF (ICOM== 9) THEN
	 GOTO 90
      ELSEIF (ICOM==10) THEN
	 GOTO 100
      ELSEIF (ICOM==11) THEN
	 GOTO 110
      ELSEIF (ICOM==12) THEN
	 GOTO 120
      ELSEIF (ICOM==13) THEN
	 GOTO 130
      ELSEIF (ICOM==14) THEN
	 GOTO 140
      ELSEIF (ICOM==15) THEN
	 GOTO 150
      ENDIF
C-----------------------------------------------------------------------
      CALL DSCAL(N,0.0,VWORK(1,1),1)
      CALL DCOPY(N,    VWORK(1,1),1,VWORK(1,3),1)
      ITER = 0
   1  ITER = ITER+1
	 IF (ITER/=1) THEN
            CLX    =   1
	    CLY    = M+4
            ICOM   =   1
            REVCOM =   5
C-----------------------------------------------------------------------
C          ASK THE USER TO PERFORM y <- D^{-1}x
C-----------------------------------------------------------------------
            RETURN
  10        CLX    = M+4
	    CLY    =   3
	    ICOM   =   2
            REVCOM =   1
C-----------------------------------------------------------------------
C          ASK THE USER TO PERFORM y <- Ax
C-----------------------------------------------------------------------
            RETURN
  20        CALL DSCAL(N,-1.0,            VWORK(1,3),1)
         ENDIF
         CALL    DAXPY(N,1.0,VWORK(1,2),1,VWORK(1,3),1)
         CLX    =   3
         CLY    = M+4
         ICOM   =   3
         REVCOM =   2
C-----------------------------------------------------------------------
C       ASK THE USER TO PERFORM y <- P^{-1}x
C-----------------------------------------------------------------------
         RETURN
  30     CLX    = M+4
         CLY    =   3
         ICOM   =   4
         REVCOM =   3
C-----------------------------------------------------------------------
C       ASK THE USER TO PERFORM y <- Dx
C-----------------------------------------------------------------------
         RETURN
  40     CLX    =   3
         CLY    =   3
         ICOM   =   5
         REVCOM =   4
C-----------------------------------------------------------------------
C       ASK THE USER TO PERFORM THE INNER PRODUCT XCOM <- y^Tx
C-----------------------------------------------------------------------
         RETURN
  50     NORMR = SQRT(XCOM)
         IF (ITER==1) NORMR0 = NORMR
         CALL DSCAL(N,1.0/NORMR,VWORK(1,3),1)
         CALL DSCAL(M+1,0.0,HWORK(1,M+1),1)
         HWORK(1,M+1) = NORMR
         I = 0
   2     I = I+1
            CLX    = I+2
	    CLY    = M+4
	    ICOM   =   6
	    REVCOM =   5
C-----------------------------------------------------------------------
C          ASK THE USER TO PERFORM y <- D^{-1}x
C-----------------------------------------------------------------------
            RETURN
  60        CLX    = M+4
            CLY    = I+3
            ICOM   =   7
            REVCOM =   1
C-----------------------------------------------------------------------
C          ASK THE USER TO PERFORM y <- Ax
C-----------------------------------------------------------------------
            RETURN
  70        CLX    = I+3
            CLY    = M+4
            ICOM   =   8
            REVCOM =   2
C-----------------------------------------------------------------------
C          ASK THE USER TO PERFORM y <- P^{-1}x
C-----------------------------------------------------------------------
            RETURN
  80        CLX    = M+4
            CLY    = I+3
            ICOM   =   9
            REVCOM =   3
C-----------------------------------------------------------------------
C          ASK THE USER TO PERFORM y <- Dx
C-----------------------------------------------------------------------
            RETURN
C-----------------------------------------------------------------------
C          MODIFIED GRAM-SCHMIDT STEP
C-----------------------------------------------------------------------
  90        KMGS = 0
   3        KMGS = KMGS+1
               CLX    = KMGS+2
               CLY    = I+3
               ICOM   =  10
               REVCOM =   4
C-----------------------------------------------------------------------
C             ASK THE USER TO PERFORM THE INNER PRODUCT XCOM <- y^Tx
C-----------------------------------------------------------------------
               RETURN
 100           HWORK(KMGS,I) = XCOM
               CALL DAXPY(N,-HWORK(KMGS,I),VWORK(1,KMGS+2),1,
     &                    VWORK(1,I+3),1)
            IF (KMGS<I) GOTO 3
            CLX    = I+3
            CLY    = I+3
            ICOM   =  11
            REVCOM =   4
C-----------------------------------------------------------------------
C          ASK THE USER TO PERFORM THE INNER PRODUCT XCOM <- y^Tx
C-----------------------------------------------------------------------
            RETURN
 110        HWORK(I+1,I) = SQRT(XCOM)
            IF (HWORK(I+1,I)/=0.0) THEN
               CALL DSCAL(N,1.0/HWORK(I+1,I),VWORK(1,I+3),1)
            ENDIF
C-----------------------------------------------------------------------
C          UPDATE H WITH PREVIOUS GIVENS ROTATIONS
C-----------------------------------------------------------------------
            DO K = 1,I-1
	       TEMP         =  HWORK(K,M+2)*HWORK(K  ,I)+
     &                         HWORK(K,M+3)*HWORK(K+1,I)
	       HWORK(K+1,I) = -HWORK(K,M+3)*HWORK(K  ,I)+
     &                         HWORK(K,M+2)*HWORK(K+1,I)
	       HWORK(K  ,I) = TEMP
            ENDDO
C-----------------------------------------------------------------------
C          COMPUTE NEXT GIVENS ROTATION AND MAKE CURRENT UPDATES
C-----------------------------------------------------------------------
            CALL GIVENS(HWORK(I,I),HWORK(I+1,I),HWORK(I,M+2),
     &                  HWORK(I,M+3))
            TEMP           =  HWORK(I,M+2)*HWORK(I,M+1)
            HWORK(I+1,M+1) = -HWORK(I,M+3)*HWORK(I,M+1)
            HWORK(I  ,M+1) = TEMP
            HWORK(I  ,  I) =  HWORK(I,M+2)*HWORK(I  ,I)+
     &                        HWORK(I,M+3)*HWORK(I+1,I)
            ERROR = ABS(HWORK(I+1,M+1))/NORMR0
C-----------------------------------------------------------------------
C          CHECK FOR CONVERGENCE AT STEP I
C-----------------------------------------------------------------------
            IF (ERROR<=TOL) THEN
               CALL DTRSV('U','N','N',I,HWORK,M+1,HWORK(1,M+1),1)
	       CALL DGEMV('N',N,I,1.0,VWORK(1,3),LDVWORK,
     &                    HWORK(1,M+1),1,0.0,VWORK(1,M+4),1)
               CALL DAXPY(N,1.0,VWORK(1,1),1,VWORK(1,M+4),1)
               CLX    = M+4
	       CLY    =   1
	       ICOM   =  12
	       REVCOM =   5
C-----------------------------------------------------------------------
C             ASK THE USER TO PERFORM y <- D^{-1}x
C-----------------------------------------------------------------------
               RETURN
 120           CLX    = M+4
               CLY    = M+4
               ICOM   =  13
               REVCOM =   4
C-----------------------------------------------------------------------
C             ASK THE USER TO PERFORM INNER PRODUCT XCOM <- y^Tx
C-----------------------------------------------------------------------
               RETURN
 130           XCOM   = SQRT(XCOM)
               ITERMAX= ITER*M+I
	       TOL    = ERROR
               IER    =  1
               REVCOM = -1
C-----------------------------------------------------------------------
C             SUCCESSFUL CONVERGENCE, EXIT SUBROUTINE GMRES
C-----------------------------------------------------------------------      
	       RETURN
            ENDIF
C-----------------------------------------------------------------------
C       INCREASE HESSENBERG FACTORIZATION?
C-----------------------------------------------------------------------
         IF (I<M) GOTO 2
C-----------------------------------------------------------------------
C       UPDATE APPROXIMATION AT THE END OF THE CYCLE
C-----------------------------------------------------------------------
         CALL DTRSV('U','N','N',M,HWORK,M+1,HWORK(1,M+1),1)
         CALL DGEMV('N',N,M,1.0,VWORK(1,3),LDVWORK,
     &              HWORK(1,M+1),1,1.0,VWORK(1,1),1)
C-----------------------------------------------------------------------
C    RESTART GMRES WITH NEW APPROXIMATION?
C-----------------------------------------------------------------------
      IF (ITER<ITERMAX) GOTO 1
C-----------------------------------------------------------------------
C    GMRES DOES NOT CONVERGE AFTER ITERMAX STEPS
C-----------------------------------------------------------------------
      CALL DCOPY(N,VWORK(1,1),1,VWORK(1,M+4),1)
      CLX    =M+4
      CLY    =  1
      ICOM   = 14
      REVCOM =  5
C-----------------------------------------------------------------------
C    ASK THE USER TO PERFORM y <- D^{-1}x
C-----------------------------------------------------------------------
      RETURN
 140  CLX    =M+4
      CLY    =M+4
      ICOM   = 15
      REVCOM =  4
C-----------------------------------------------------------------------
C    ASK THE USER TO PERFORM INNER PRODUCT XCOM <- y^Tx
C-----------------------------------------------------------------------
      RETURN
 150  XCOM   = SQRT(XCOM)
      ITERMAX= (ITER-1)*M+I
      TOL    = ERROR
      IER    =  0
      REVCOM = -1
C-----------------------------------------------------------------------
C    NO CONVERGENCE, EXIT SUBROUTINE GMRES
C-----------------------------------------------------------------------      
      RETURN
C-----------------------------------------------------------------------
      END SUBROUTINE GMRES
C-----------------------------------------------------------------------
