       SUBROUTINE RICHD(N,ITERMAX,TOL,VWORK,LDVWORK,
     &                  REVCOM,CLX,CLY,XCOM,IER)
C
C SOLVE Ax=b USING THE PRECONDITIONNED RICHARDSON METHOD
C
C ITERMAX:    ON ENTRY: MAXIMUM NUMBER OF ITERATIONS
C             ON EXIT : NUMBER OF MATRIX-VECTOR PRODUCTS
C
C TOL:        ON ENTRY: TOLERANCE FOR ITERATION
C             ON EXIT : NORM OF THE RESIDUAL ERROR
C
C VWORK:      VWORK(:,1): ON EXIT:  CONTAINS THE APPROXIMATE SOLUTION x
C             VWORK(:,2): ON ENTRY: CONTAINS THE RIGHT-HAND SIDE b
C             VWORK(:,3),VWORK(:,4): USED AS WORKING VECTORS AND FOR REVERSE COMMUNICATION
C             ON EXIT VWORK(:,4) CONTAINS Dx
C
C REVCOM:     ON ENTRY REVCOM =  0
C             ON EXIT  REVCOM =  1 MATRIX VECTOR PRODUCT BY A
C                      REVCOM =  2 MATRIX VECTOR PRODUCT BY P^{-1}
C                      REVCOM =  3 MATRIX VECTOR PRODUCT BY D
C                      REVCOM =  4 INNER PRODUCT
C                      REVCOM = -1 EXIT
C
C IER: INTEGER (OUTPUT) IER =  1 SUCCESSFUL CONVERGENCE
C                       IER =  0 NO CONVERGENCE AFTER ITERMAX ITERATIONS
C
C CAUTION: ONLY THE INPUT ONES CAN BE MODIFIED DURING OUT OF CORE COMPUTATIONS
C
C-----------------------------------------------------------------------
C    DECLARATIONS
C-----------------------------------------------------------------------
      IMPLICIT  NONE
C    SUBROUTINE VARIABLES
      INTEGER, INTENT(IN   ) :: N       ! SIZE OF THE PROBLEM
      INTEGER, INTENT(INOUT) :: ITERMAX
      REAL   , INTENT(INOUT) :: TOL
      INTEGER, INTENT(IN   ) :: LDVWORK ! LEADING DIMENSION OF VWORK, N IN GENERAL
      REAL   , INTENT(INOUT) :: VWORK(LDVWORK,4)
      INTEGER, INTENT(INOUT) :: REVCOM
      INTEGER, INTENT(  OUT) :: CLX     ! COLUMN OF  INPUT IN VWORK FOR REVERSE COMMUNICATION
      INTEGER, INTENT(  OUT) :: CLY     ! COLUMN OF OUTPUT IN VWORK FOR REVERSE COMMUNICATION
      REAL   , INTENT(INOUT) :: XCOM    ! ON ENTRY CONTAINS INNER PRODUCT WHEN REVCOM = 4
      INTEGER, INTENT(  OUT) :: IER
C    LOCAL SAVED VARIABLES
      INTEGER, SAVE          :: ICOM,ITER
      REAL   , SAVE          :: ERROR,NORMR0
C-----------------------------------------------------------------------
C    BEGINNING OF SUBROUTINE RICHD
C-----------------------------------------------------------------------
      IF (REVCOM==0) ICOM = REVCOM
      IF     (ICOM==1) THEN
         GOTO 10
      ELSEIF (ICOM==2) THEN
         GOTO 20
      ELSEIF (ICOM==3) THEN
         GOTO 30
      ELSEIF (ICOM==4) THEN
         GOTO 40
      ELSEIF (ICOM==5) THEN
         GOTO 50
      ELSEIF (ICOM==6) THEN
         GOTO 60
      ELSEIF (ICOM==7) THEN
         GOTO 70
      ELSEIF (ICOM==8) THEN
         GOTO 80
      ENDIF
C-----------------------------------------------------------------------
      CALL DSCAL(N,0.0,VWORK(1,1),1)
      CALL DCOPY(N,    VWORK(1,1),1,VWORK(1,3),1)
      ITER = 0
   1  ITER = ITER+1
         IF (ITER/=1) THEN
            CLX    =  1
	    CLY    =  3
            ICOM   =  1
            REVCOM =  1
C-----------------------------------------------------------------------
C          ASK THE USER TO PERFORM y <- Ax
C-----------------------------------------------------------------------
            RETURN
  10        CALL DSCAL(N,-1.0,            VWORK(1,3),1)
         ENDIF  
         CALL    DAXPY(N,1.0,VWORK(1,2),1,VWORK(1,3),1)
         CLX    =  3
         CLY    =  4
         ICOM   =  2
         REVCOM =  2
C-----------------------------------------------------------------------
C       ASK THE USER TO PERFORM y <- P^{-1}x
C-----------------------------------------------------------------------
         RETURN
  20     CLX    =  4
         CLY    =  3
         ICOM   =  3
         REVCOM =  3
C-----------------------------------------------------------------------
C       ASK THE USER TO PERFORM y <- Dx
C-----------------------------------------------------------------------
         RETURN
  30     CLX    =  3
         CLY    =  3 
         ICOM   =  4
         REVCOM =  4
C-----------------------------------------------------------------------
C       ASK THE USER TO PERFORM INNER PRODUCT XCOM <- y^Tx
C-----------------------------------------------------------------------
         RETURN
  40     IF (ITER==1) NORMR0 = SQRT(XCOM)
C-----------------------------------------------------------------------
C       UPDATE APPROXIMATION
C-----------------------------------------------------------------------
         CALL DAXPY(N,1.0,VWORK(1,4),1,VWORK(1,1),1)
         ERROR = SQRT(XCOM)/NORMR0
         IF (ERROR<=TOL) THEN
            CLX    =  1
            CLY    =  4
	    ICOM   =  5
            REVCOM =  3
C-----------------------------------------------------------------------
C          ASK THE USER TO PERFORM y <- Dx
C-----------------------------------------------------------------------
            RETURN
  50        CLX    =  4
	    CLY    =  4
            ICOM   =  6
            REVCOM =  4
C-----------------------------------------------------------------------
C          ASK THE USER TO PERFORM INNER PRODUCT XCOM <- y^Tx
C-----------------------------------------------------------------------
            RETURN
  60        XCOM   = SQRT(XCOM)
            ITERMAX= ITER
            TOL    = ERROR
            IER    =  1
            REVCOM = -1
C-----------------------------------------------------------------------
C          SUCCESSFUL CONVERGENCE, EXIT SUBROUTINE RICHD
C-----------------------------------------------------------------------      
            RETURN
         ENDIF
C-----------------------------------------------------------------------
C    ITERATE?
C-----------------------------------------------------------------------
      IF (ITER<ITERMAX) GOTO 1
C-----------------------------------------------------------------------
C    RICHARDSON DID NOT CONVERGE AFTER ITERMAX ITERATIONS
C-----------------------------------------------------------------------
      CLX    =  1
      CLY    =  4
      ICOM   =  7
      REVCOM =  3
C-----------------------------------------------------------------------
C    ASK THE USER TO PERFORM y <- Dx
C-----------------------------------------------------------------------
      RETURN
  70  CLX    =  4
      CLY    =  4
      ICOM   =  8
      REVCOM =  4
C-----------------------------------------------------------------------
C    ASK THE USER TO PERFORM INNER PRODUCT XCOM <- y^Tx
C-----------------------------------------------------------------------
      RETURN
  80  XCOM   = SQRT(XCOM)
      ITERMAX= ITER
      TOL    = ERROR
      IER    =  0
      REVCOM = -1
C-----------------------------------------------------------------------
C    NO CONVERGENCE, EXIT SUBROUTINE RICHD
C-----------------------------------------------------------------------
      RETURN
C-----------------------------------------------------------------------
      END SUBROUTINE RICHD
C-----------------------------------------------------------------------
