      PROGRAM DAHLQUIST
C-----------------------------------------------------------------------
C    DRIVER FOR DAHLQUIST PROBLEM USING SPARK3
C           y'=lambda*y with lambda=-10.3 and y(0)=10
C-----------------------------------------------------------------------
C    DECLARATIONS
C-----------------------------------------------------------------------
      IMPLICIT  NONE
      INCLUDE '../../INC/sp.h'
      INCLUDE '../../INC/mpipath.h'
C    PARAMETERS
      INTEGER, PARAMETER     :: LRWORK = 300 ! 160
      INTEGER, PARAMETER     :: LIWORK = 200 ! 141
      INTEGER, PARAMETER     :: LRPAR  = 1
      INTEGER, PARAMETER     :: LIPAR  = 0
      INTEGER, PARAMETER     :: NM = 1
      INTEGER, PARAMETER     :: NY = 1
      INTEGER, PARAMETER     :: NU = NY
      INTEGER, PARAMETER     :: NV = 0
C    GLOBAL VARIABLES
      INTEGER :: I,IDEVCN,IDEVVECCN,IDFYCNM(NM),IDFYVECCNM(NM),IDID,IOUT
      INTEGER :: IPAR(LIPAR),ITOLY,IWORK(LIWORK),LDFYM(NM),LDEV
      INTEGER :: ME(NM),NACCEPT,NDEV,NDFY,NECN,NFCN,NJACDEC,NJACSOL
      INTEGER :: NLINIT,NMATVECPROD,NNEWT,NPVNDEC,NPVNSOL,NREJECTERR
      INTEGER :: NRLOOP,NS,NSTEP,NTOL,NTOLDF,NTOLMN,NTOLMX
      REAL    :: ATOLY,ERRTOL,ERRTOLI,FACNORMNY,H,RPAR(LRPAR)
      REAL    :: RTOLY,RWORK(LRWORK),SCALYINVI,T,TEND,TOLFC,TOLST
      REAL    :: Y(NY),DLAMBDA,T0,Y0,Y1
      REAL    (sp) :: TARRAY(2)
      EXTERNAL DEVCN,DEVVECCN,DFYCNM,DFYVECCNM,ECN,FCNM,PRDECVN,PRSOLVN
      EXTERNAL PRECDEC,PRECSOL,SOLOUT
      INTEGER                :: MYID,COMM,ERRMPI,IS,ITEMP
C-----------------------------------------------------------------------
C    BEGINNING OF PROGRAM SIMPLE
C-----------------------------------------------------------------------

      CALL MPI_INIT(ERRMPI)
      CALL MPI_COMM_RANK(MPI_COMM_WORLD,MYID,ERRMPI)
      COMM=MPI_COMM_WORLD
      IS=MYID+1
      ITEMP=0

      FACNORMNY=1.0/SQRT(REAL(NY))

C ---  LOBATTO IIIA  -----
c      ME(1)=1
C ---  LOBATTO IIIB  -----
c      ME(1)=2
C ---  LOBATTO IIIC  -----
c      ME(1)=3
C ---  LOBATTO IIIC* -----
c      ME(1)=4
C ---  LOBATTO IIID  -----
c      ME(1)=5
C ---  RADAU IIA -----
      ME(1)=11
C ---  GAUSS -----
c      ME(1)=21

c ----- DATA FILE -----
      IF (IS.EQ.1) THEN
      OPEN( 8,FILE='res_spark3')
      REWIND 8
      ENDIF

C --- LOOP FOR DIFFERENT TOLERANCES
      NTOLMN=3
c      NTOLMX=6
      NTOLMX=12
      NTOLDF=1
      NRLOOP=(NTOLMX-NTOLMN)*NTOLDF
      TOLST=0.1**NTOLMN
      TOLFC=0.1**(1.0/REAL(NTOLDF))

      DO NTOL=0,NRLOOP

C ----- INITIAL VALUES -----
      CALL INIT(T,NY,Y,RPAR,IPAR)
      T0=T
      Y0=Y(1)

C ----- ENDPOINT OF INTEGRATION -----
      TEND=20.0

C-----------------------------------------------------------------------
C    PREPARATIONS FOR SPARK3
C-----------------------------------------------------------------------

C ----- REQUIRED TOLERANCE -----
      RTOLY=TOLST
      ATOLY=RTOLY

C ----- BOTH RTOLY AND ATOLY ARE SCALARS -----
      ITOLY=0

C ----- INITIAL STEP SIZE -----
c      H=1.E-6
c      H=1.E-5
      H=1.E-4
c      H=1.E-3
c      H=1.E-2
c      H=1.E-1

C ----- OUTPUT ROUTINE IS NOT USED DURING INTEGRATION -----
      IOUT=0

C ----- SET DEFAULT VALUES -----
      DO I=1,128
         IWORK(I)=-1
      ENDDO
      DO I=1,64
         RWORK(I)=-1.0
      ENDDO

C     MAXIMUM NUMBER OF ALLOWED STEPS.
C default 
c      IWORK(1)=100000

C     MINIMUM NUMBER OF INEXACT APPROXIMATE SIMPLIFIED NEWTON
C     ITERATIONS AT EACH STEP.
C default 
c      IWORK(2)=1

C     MAXIMUM NUMBER OF INEXACT APPROXIMATE SIMPLIFIED NEWTON
C     ITERATIONS AT EACH STEP.
C default 
c      IWORK(3)=7
      IWORK(3)=20

C     SWITCH FOR THE STARTING VALUES OF THE INEXACT APPROXIMATE
C     SIMPLIFIED NEWTON ITERATIONS.
C     THEY ARE SET TO ZERO
c      IWORK(4)=0
C default
C     THEY ARE SET WITH THE DENSE OUTPUT FORMULA
c      IWORK(4)=1

C     SWITCH INDICATING IF ONE SET OF SPARK COEFFICIENTS USED FOR
C     THE DIFFERENTIAL EQUATION e(t,y)'=... IS NOT STIFFLY ACCURATE.
C default
C     NV<=0 OR ONLY STIFFLY ACCURATE COEFFICIENTS ARE USED FOR e(t,y)'
c      IWORK(5)=0
C               NON-STIFFLY ACCURATE COEFFICIENTS ARE USED FOR e(t,y)'
c      IWORK(5)=1

C     SWITCH FOR THE PRECONDITIONED LINEAR ITERATIVE METHOD.
C     NO ITERATIONS (LEADS TO "FIXED-POINT" ITERATIONS)
c      IWORK(6)=0
C     RICHARDSON
      IWORK(6)=1
C default
C     GMRES(IWORK(29))
c      IWORK(6)=2

C     NUMBER OF MATRICES H_i TO DECOMPOSE
C default
C      IWORK(7)=1
CMPI
      IWORK(7)=3
CMPIend

C     SWITCH FOR THE STEPSIZE CONTROL STRATEGY.
C     CLASSICAL VARIABLE STEPSIZE CONTROL
c      IWORK(8)=0
C     CONSTANT STEPSIZES
c      IWORK(8)=1
C default
C     PREDICTIVE VARIABLE STEPSIZE CONTROL OF GUSTAFSSON
c      IWORK(8)=2

C     LENGTH OF EACH REAL    WORKSPACE  BHI USED FOR THE PRECONDITIONER
C     IN SUBROUTINES PRECDEC AND PRECSOL.
C default
c      IWORK(9)=1
      IWORK(9)=NY*NY

C     LENGTH OF EACH INTEGER WORKSPACE IBHI USED FOR THE PRECONDITIONER
C     IN SUBROUTINES PRECDEC AND PRECSOL.
C default
c      IWORK(11)=1
      IWORK(11)=NY

C     LENGTH OF      REAL    WORKSPACE DFY  USED FOR THE PRECONDITIONER
C     IN SUBROUTINES DFYCN AND PRECDEC.
C default
c      LDFYM(1)=1
      LDFYM(1)=NY*NY*NM

C     LENGTH OF      REAL    WORKSPACE DEV  USED FOR THE PRECONDITIONERS
C     IN SUBROUTINES DEVCN, PRECDEC, AND PRDECVN.
C default
c      LDEV=1
      LDEV=NV

C     LENGTH OF REAL    WORKSPACE  VP USED FOR THE PRECONDITIONER TO Ev
C     IN SUBROUTINES PRDECVN AND PRSOLVN.
C default 
c      IWORK(10)=1

C     LENGTH OF INTEGER WORKSPACE IVP USED FOR THE PRECONDITIONER TO Ev
C     IN SUBROUTINES PRDECVN AND PRSOLVN.
C default
c      IWORK(12)=1

C     SWITCH TO COMPUTE QUANTITIES NEEDED FOR THE APPROXIMATE JACOBIANS Fmy.
C default
C     Fmy ARE COMPUTED INTERNALLY BY USING NUMERICAL
C     FINITE DIFFERENCES APPROXIMATIONS TO PARTIAL DERIVATIVES. 
C     SUBROUTINE DFYCN IS NEVER CALLED. SUPPLY A DUMMY SUBROUTINE.
c      IDFYCNM(1)=0
C     QUANTITIES NEEDED FOR Fmy ARE COMPUTED BY SUBROUTINE DFYCN.
      IDFYCNM(1)=1
C     NO Fmy ARE COMPUTED, SUBROUTINE DFYCN IS NEVER CALLED.
C     SUPPLY A DUMMY SUBROUTINE.
c      IDFYCNM(1)=2

C     SWITCH TO COMPUTE APPROXIMATE JACOBIANS Fmy-VECTOR PRODUCTS FOR THE
C     PRECONDITIONED LINEAR ITERATIVE METHOD.
C default
C     STANDARD PRODUCT USING Fmy EXPLICITLY
c      IDFYVECCNM(1)=0
C     Fmy-FREE PRODUCT USING FINITE DIFFERENCES
c      IDFYVECCNM(1)=1
C     USER-SUPPLIED SUBROUTINE DFYVECCN
      IDFYVECCNM(1)=2

C     SWITCH TO COMPUTE APPROXIMATE JACOBIAN Ev-VECTOR PRODUCTS FOR THE
C     PRECONDITIONED LINEAR ITERATIVE METHOD.
C default
C     STANDARD PRODUCT USING Ev EXPLICITLY
c      IDEVVECCN=0
C     Ev-FREE PRODUCT USING FINITE DIFFERENCES
c      IDEVVECCN=1
C     USER-SUPPLIED SUBROUTINE DEVVECCN
      IDEVVECCN=2

C     SWITCH TO COMPUTE QUANTITIES NEEDED FOR THE APPROXIMATE JACOBIAN Ev.
C default
C     Ev IS COMPUTED INTERNALLY BY USING NUMERICAL
C     FINITE DIFFERENCES APPROXIMATIONS TO PARTIAL DERIVATIVES. 
C     SUBROUTINE DEVCN IS NEVER CALLED. SUPPLY A DUMMY SUBROUTINE.
c      IDEVCN=0
C     QUANTITIES NEEDED FOR Ev ARE COMPUTED BY SUBROUTINE DEVCN.
      IDEVCN=1
C     NV<=0 OR NO Ev IS COMPUTED, SUBROUTINE DEVCN
C     IS NEVER CALLED. SUPPLY A DUMMY SUBROUTINE.
c      IDEVCN=2

C     MAXIMUM NUMBER OF ITERATIONS FOR THE PRECONDITIONED ITERATIVE METHOD
      IWORK(28)=1
c      IWORK(28)=2
c      IWORK(28)=3
c      IWORK(28)=4
C default
c      IWORK(28)=20

      IF     (IWORK(6)==0) THEN
C ----- NO INITIALIZATION NECESSARY                      -----

      ELSEIF (IWORK(6)==1) THEN
C ----- INITIALIZATION OF SOME PARAMETERS FOR RICHARDSON -----

      ELSEIF (IWORK(6)==2) THEN
C ----- INITIALIZATION OF SOME PARAMETERS FOR GMRES -----
C        MAXIMUM DIMENSION OF THE KRYLOV SUBSPACE (RESTART PARAMETER)
c         IWORK(29)=1
c         IWORK(29)=2
c         IWORK(29)=3
c         IWORK(29)=4
c         IWORK(29)=5
C    default
c         IWORK(29)=10
      ENDIF


C     ROUNDING UNIT, SMALLEST NUMBER v  ON THE COMPUTER SATISFYING 1.0+v>1.0.
C default
c      RWORK(1)=EPSILON(1.0)/2.0

C     THE SAFETY FACTOR IN STEPSIZE PREDICTION.
C default
c      RWORK(2)=0.9

C     REFERENCE RATE OF CONVERGENCE FOR THE INEXACT APPROXIMATE SIMPLIFIED NEWTON METHOD
c      RWORK(3)=0.5
c      RWORK(3)=0.2
C default
c      RWORK(3)=0.15
c      RWORK(3)=0.1
c      RWORK(3)=0.01

C     STOPPING CRITERION FOR THE INEXACT APPROXIMATE SIMPLIFIED NEWTON METHOD, USUALLY CHOSEN <1.
c      RWORK(4)=1.0
c      RWORK(4)=0.3
c      RWORK(4)=0.03
C default
c      RWORK(4)=0.01
c      RWORK(4)=0.003
c      RWORK(4)=0.00001

C     STOPPING CRITERION FACTOR FOR THE PRECONDITIONED LINEAR ITERATIVE METHOD.
c      RWORK(5)=0.3
C default
c      RWORK(5)=0.1
c      RWORK(5)=0.01
c      RWORK(5)=0.001
c      RWORK(5)=0.0001
c      RWORK(5)=0.000001
c      RWORK(5)=0.0000001

C     MINIMUM ABSOLUTE STEPSIZE LENGTH.
C default
c      RWORK(6)=ABS(SQRT(RWORK(1)))

C     MAXIMUM ABSOLUTE STEPSIZE LENGTH.
C default
c      RWORK(7)=ABS(TEND-T)

C     PARAMETERS FOR STEPSIZE SELECTION. THE NEW STEPSIZE IS CHOSEN SUBJECT
C     TO THE RESTRICTION RWORK(8) <= HNEW/H <= RWORK(9)
C default
c      RWORK(8)=0.2
c      RWORK(8)=0.8
C default
c      RWORK(9)=8.0
c      RWORK(9)=1.2

C     PARAMETERS FOR THE EVALUATION AND DECOMPOSITION OF A NEW APPROXIMATE JACOBIAN
c      RWORK(12)=0.2
C default
c      RWORK(12)=0.1
c      RWORK(12)=0.01
c      RWORK(12)=0.001
c      RWORK(12)=0.0

c      RWORK(13)=0.2
C default
c      RWORK(13)=0.1
c      RWORK(13)=0.03
c      RWORK(13)=0.01
c      RWORK(13)=0.001
c      RWORK(13)=0.0

C ----- CALL OF THE SUBROUTINE SPARK3 -----
      IF (IS.EQ.1) THEN
      CALL DTIME(TARRAY)
      ENDIF
      NS=3
      CALL      PSPARK3(NS,NM,ME,T,TEND,NY,NU,NV,Y,
     &                  FCNM,DFYCNM,IDFYCNM,DFYVECCNM,IDFYVECCNM,LDFYM,
     &                  ECN ,DEVCN ,IDEVCN ,DEVVECCN ,IDEVVECCN ,LDEV ,
     &                  PRECDEC,PRECSOL,
     &                  PRDECVN,PRSOLVN,
     &                  H,ITOLY,RTOLY,ATOLY,
     &                  IOUT,SOLOUT,
     &                  LRWORK,RWORK,LIWORK,IWORK,
     &                  LRPAR ,RPAR ,LIPAR ,IPAR ,
     &                  IDID,
     &                  MYID,COMM)
      IF (IS.EQ.1) THEN
      CALL DTIME(TARRAY)

C ----- PRINT SOLUTION -----
      WRITE(8,9921) Y(1)
 9921 FORMAT(1X,E22.16)

      ERRTOL=0.0
      SCALYINVI=1.0/(ATOLY+RTOLY*ABS(Y(I)))
      DLAMBDA=RPAR(1)
      Y1=Y0*EXP(DLAMBDA*(TEND-T0))
      WRITE (6,*) 'EXACT     SOLUTION: Y(1)=',Y1
      WRITE (6,*) 'NUMERICAL SOLUTION: Y(1)=',Y(1)
      ERRTOL=ABS(Y1-Y(1))*SCALYINVI
      ERRTOL=ERRTOL*FACNORMNY
      WRITE (6,9922) ERRTOL
      WRITE (8,9922) ERRTOL
 9922 FORMAT(' ERRTOL= ' ,E22.16)
C ----- PRINT STATISTICS -----
      WRITE(8,*) 'ELAPSED TIME=',TARRAY(1)
      WRITE(6,*) 'ELAPSED TIME=',TARRAY(1)
      WRITE(6,90) ATOLY,RTOLY
      WRITE(8,90) ATOLY,RTOLY
 90   FORMAT(' ATOLY=',E8.2,' RTOLY=',E8.2)
      NFCN       =IWORK(30)
      NECN       =IWORK(31)
      NDFY       =IWORK(32)
      NDEV       =IWORK(33)
      NSTEP      =IWORK(34)
      NACCEPT    =IWORK(35)
      NREJECTERR =IWORK(36)
      NLINIT     =IWORK(37)
      NNEWT      =IWORK(38)
      NJACDEC    =IWORK(39)
      NJACSOL    =IWORK(40)
      NPVNDEC    =IWORK(41)
      NPVNSOL    =IWORK(42)
      NMATVECPROD=IWORK(43)
      WRITE(6,91) NFCN,NECN
      WRITE(8,91) NFCN,NECN
 91   FORMAT('    FCN=',I6,'      ECN=',I6)
      WRITE(6,92) NDFY,NDEV
      WRITE(8,92) NDFY,NDEV
 92   FORMAT('   NDFY=',I6,'     NDEV=',I6)
      WRITE(6,93) NSTEP,NACCEPT,NREJECTERR
      WRITE(8,93) NSTEP,NACCEPT,NREJECTERR
 93   FORMAT('  NSTEP=',I6,'  NACCEPT=',I6,'  NREJECTERR=',I6)
      WRITE(6,94) NLINIT,NNEWT
      WRITE(8,94) NLINIT,NNEWT
 94   FORMAT(' NLINIT=',I6,'    NNEWT=',I6)
      WRITE(6,95) NJACDEC,NJACSOL
      WRITE(8,95) NJACDEC,NJACSOL
 95   FORMAT('NJACDEC=',I6,'  NJACSOL=',I6)
      WRITE(6,96) NPVNDEC,NPVNSOL,NMATVECPROD
      WRITE(8,96) NPVNDEC,NPVNSOL,NMATVECPROD
 96   FORMAT('NPVNDEC=',I6,'  NPVNSOL=',I6,' NMATVECPROD=',I6)
      WRITE(8,*)

      IF (TARRAY(1)>500.0_sp) ITEMP = 1
      ENDIF
      CALL MPI_BCAST(ITEMP,1,MPI_INTEGER,0,COMM,ERRMPI)
      IF (ITEMP.EQ.1) GOTO 1234
C -------- NEW TOLERANCE ---
      TOLST=TOLST*TOLFC
      ENDDO
 1234 CONTINUE  
      IF (IS.EQ.1) THEN
      CLOSE(8)
      ENDIF

      CALL MPI_FINALIZE(ERRMPI)

      STOP
C-----------------------------------------------------------------------
      END PROGRAM DAHLQUIST
C-----------------------------------------------------------------------
